import { node_path as path } from '../runtime/mod.ts';
import Option, { None, Some } from '../option.ts';

// ----------------------------------------------------------------------------
// File-related types
// ----------------------------------------------------------------------------

export enum FileLang {
	TypeScript = 'TypeScript',
	JavaScript = 'JavaScript',
	CSS = 'CSS',
	Plain = 'Plain Text',
}

export interface HighlightingOptions {
	numbers: boolean;
	strings: boolean;
}

interface IFileType {
	readonly name: FileLang;
	readonly singleLineComment: Option<string>;
	readonly keywords1: string[];
	readonly keywords2: string[];
	readonly hlOptions: HighlightingOptions;
	get flags(): HighlightingOptions;
	get primaryKeywords(): string[];
	get secondaryKeywords(): string[];
}

/**
 * The base class for File Types
 */
export abstract class AbstractFileType implements IFileType {
	public readonly name: FileLang = FileLang.Plain;
	public readonly singleLineComment = None;
	public readonly keywords1: string[] = [];
	public readonly keywords2: string[] = [];
	public readonly hlOptions: HighlightingOptions = {
		numbers: false,
		strings: false,
	};

	get flags(): HighlightingOptions {
		return this.hlOptions;
	}

	get primaryKeywords(): string[] {
		return this.keywords1;
	}

	get secondaryKeywords(): string[] {
		return this.keywords2;
	}
}

// ----------------------------------------------------------------------------
// FileType implementations
// ----------------------------------------------------------------------------
const defaultHighlightOptions: HighlightingOptions = {
	numbers: true,
	strings: true,
};

class JavaScriptFile extends AbstractFileType {
	public readonly name: FileLang = FileLang.JavaScript;
	public readonly singleLineComment = Some('//');
	public readonly keywords1 = [
		'break',
		'case',
		'catch',
		'class',
		'const',
		'continue',
		'debugger',
		'default',
		'delete',
		'do',
		'else',
		'export',
		'extends',
		'false',
		'finally',
		'for',
		'function',
		'if',
		'import',
		'in',
		'instanceof',
		'new',
		'null',
		'return',
		'super',
		'switch',
		'this',
		'throw',
		'true',
		'try',
		'typeof',
		'var',
		'void',
		'while',
		'with',
		'let',
		'static',
		'yield',
		'await',
	];
	public readonly keywords2 = [
		'arguments',
		'as',
		'async',
		'eval',
		'from',
		'get',
		'of',
		'set',
	];
	public readonly hlOptions: HighlightingOptions = {
		...defaultHighlightOptions,
	};
}

class TypeScriptFile extends JavaScriptFile {
	public readonly name: FileLang = FileLang.TypeScript;
	public readonly keywords1 = [
		'break',
		'case',
		'catch',
		'class',
		'const',
		'continue',
		'debugger',
		'default',
		'delete',
		'do',
		'else',
		'export',
		'extends',
		'false',
		'finally',
		'for',
		'function',
		'if',
		'import',
		'in',
		'instanceof',
		'new',
		'null',
		'return',
		'super',
		'switch',
		'this',
		'throw',
		'true',
		'try',
		'typeof',
		'var',
		'void',
		'while',
		'with',
		'let',
		'static',
		'yield',
		'await',
	];
	public readonly keywords2 = [
		'arguments',
		'as',
		'async',
		'eval',
		'from',
		'get',
		'of',
		'set',
		// Typescript-specific
		'keyof',
		'interface',
		'enum',
		'public',
		'protected',
		'private',
		'string',
		'number',
		'boolean',
		'any',
		'unknown',
		'type',
	];
}

class CSSFile extends AbstractFileType {
	public readonly name: FileLang = FileLang.CSS;
	public readonly singleLineComment = None;
	public readonly hlOptions: HighlightingOptions = {
		...defaultHighlightOptions,
	};
}

// ----------------------------------------------------------------------------
// External interface
// ----------------------------------------------------------------------------

export class FileType extends AbstractFileType {
	static #fileTypeMap = new Map([
		['.css', CSSFile],
		['.js', JavaScriptFile],
		['.jsx', JavaScriptFile],
		['.mjs', JavaScriptFile],
		['.ts', TypeScriptFile],
		['.tsx', TypeScriptFile],
	]);

	public static default(): FileType {
		return new FileType();
	}

	public static from(filename: string): FileType {
		const ext = path.extname(filename);
		const type = FileType.#fileTypeMap.get(ext) ?? FileType;

		return new type();
	}
}
